from attrs import define, field, validators
from partial_date import PartialDate

from illallangi.rdf.converters.partial_date import to_partial_date


@define(kw_only=True)
class ResidenceKey:
    # Natural Keys

    label: str = field(
        validator=[
            validators.instance_of(str),
            validators.matches_re(r"^[\w ]{1,63}$"),
        ],
    )


@define(kw_only=True)
class Residence(ResidenceKey):
    country: str = field(
        validator=[
            validators.instance_of(str),
            validators.matches_re(r"^[\w ]{1,63}$"),
        ],
    )

    finish: PartialDate | None = field(
        converter=to_partial_date,
        default="",
        validator=[
            validators.instance_of(PartialDate | None),
        ],
    )

    locality: str = field(
        validator=[
            validators.instance_of(str),
            validators.matches_re(r"^[\w ]{1,63}$"),
        ],
    )

    open_location_code: str = field(
        validator=[
            validators.instance_of(str),
            validators.matches_re(
                r"^[23456789CFGHJMPQRVWX]{8}\+[23456789CFGHJMPQRVWX]{2,}$"
            ),
        ],
    )

    postal_code: str = field(
        validator=[
            validators.instance_of(str),
            validators.matches_re(r"^\d{4}$"),
        ],
    )

    region: str = field(
        validator=[
            validators.instance_of(str),
            validators.matches_re(r"^[\w ]{1,63}$"),
        ],
    )

    start: PartialDate | None = field(
        converter=to_partial_date,
        default="",
        validator=[
            validators.instance_of(PartialDate | None),
        ],
    )

    street: str = field(
        validator=[
            validators.instance_of(str),
            validators.matches_re(r"^[\w /,]{1,63}$"),
        ],
    )



from illallangi.rdf.models.aircraft import Aircraft, AircraftKey
from illallangi.rdf.models.airline import Airline, AirlineKey
from illallangi.rdf.models.airport import Airport, AirportKey
from illallangi.rdf.models.alliance import Alliance, AllianceKey
from illallangi.rdf.models.course import Course, CourseKey
from illallangi.rdf.models.manufacturer import (
    Manufacturer,
    ManufacturerKey,
)
from illallangi.rdf.models.residence import Residence, ResidenceKey

__all__ = [
    "Aircraft",
    "AircraftKey",
    "Airline",
    "AirlineKey",
    "Airport",
    "AirportKey",
    "Alliance",
    "AllianceKey",
    "Course",
    "CourseKey",
    "Manufacturer",
    "ManufacturerKey",
    "Residence",
    "ResidenceKey",
]



from collections.abc import Generator
from typing import Any

from illallangi.rdf.models import Airline, Alliance


class AirlinesNotFoundError(Exception):
    def __init__(
        self,
        airline_iata: list[str],
    ) -> None:
        self.airline_iata = airline_iata
        msg = f"Airline(s) not found: {', '.join(airline_iata)}"
        super().__init__(msg)


class AirlineMixin:
    def get_airlines(
        self,
        *_args: list[Any],
        airline_iata: list[str] | None = None,
        **_kwargs: dict[str, Any],
    ) -> Generator[Airline, Any, list | None]:
        if airline_iata is None:
            return []
        airline_iata = [i.upper() for i in airline_iata]

        query = f"""
    SELECT ?label ?iata ?icao ?alliance ?dominant_color WHERE {{
        VALUES (?value) {{ ( "{'" ) ( "'.join(airline_iata)}" ) }}
        ?href ip:airlineIataCode ?value.
        ?href rdfs:label ?label .
        ?href ip:airlineIataCode ?iata .
        OPTIONAL {{ ?href ip:airlineIcaoCode ?icao . }}
        OPTIONAL {{
            ?href ip:memberOfAirlineAlliance ?allianceHref .
            ?allianceHref rdfs:label ?alliance .
        }}
        OPTIONAL {{ ?href ip:dominantColor ?dominant_color . }}
        ?href a ic:airline .
    }}
    """

        found = []

        for airline in self.query(
            query=query,
        ):
            found.append(airline["iata"])
            yield Airline(
                **{
                    **airline,
                    "alliance": Alliance(name=str(airline["alliance"]))
                    if "alliance" in airline
                    else None,
                },
            )

        not_found = [i for i in airline_iata if i not in found]
        if not_found:
            raise AirlinesNotFoundError(not_found)



import diffsync


class Airport(
    diffsync.DiffSyncModel,
):
    _modelname = "Airport"
    _identifiers = ("iata",)
    _attributes = (
        "icao",
        "label",
    )

    iata: str

    icao: str
    label: str

    @classmethod
    def create(
        cls,
        adapter: diffsync.Adapter,
        ids: dict,
        attrs: dict,
    ) -> "Airport":
        raise NotImplementedError

    def update(
        self,
        attrs: dict,
    ) -> "Airport":
        raise NotImplementedError

    def delete(
        self,
    ) -> "Airport":
        raise NotImplementedError



from attrs import define, field, validators

from illallangi.rdf.models.alliance import Alliance


@define(kw_only=True)
class AirlineKey:
    # Natural Keys

    iata: str = field(
        validator=[
            validators.instance_of(str),
            validators.matches_re(r"^[0-9A-Z]{2}$"),
        ],
    )


@define(kw_only=True)
class Airline(AirlineKey):
    # Fields

    alliance: Alliance | None = field(
        validator=[
            validators.instance_of(Alliance | None),
        ],
    )

    dominant_color: str = field(
        validator=[
            validators.instance_of(str),
            validators.matches_re(r"^#[0-9a-f]{6}$"),
        ],
    )

    icao: str | None = field(
        default=None,
        validator=[
            validators.instance_of(str | None),
            # validators.matches_re(r"^[0-9A-Z]{3}$"),
        ],
    )

    label: str = field(
        validator=[
            validators.instance_of(str),
        ],
    )



from attrs import define, field, validators


@define(kw_only=True)
class AirportKey:
    # Natural Keys

    iata: str = field(
        validator=[
            validators.instance_of(str),
            validators.matches_re(r"^[A-Z]{3}$"),
        ],
    )


@define(kw_only=True)
class Airport(AirportKey):
    # Fields

    icao: str = field(
        validator=[
            validators.instance_of(str),
            validators.matches_re(r"^[A-Z]{4}$"),
        ],
    )

    label: str = field(
        validator=[
            validators.instance_of(str),
        ],
    )



from attrs import define, field, validators
from partial_date import PartialDate

from illallangi.rdf.converters.partial_date import to_partial_date
from illallangi.rdf.models.airline import AirlineKey
from illallangi.rdf.models.manufacturer import ManufacturerKey


@define(kw_only=True)
class AircraftKey:
    # Natural Keys

    manufacturer: ManufacturerKey = field(
        validator=[
            validators.instance_of(ManufacturerKey),
        ],
    )

    msn: str = field(
        validator=[
            validators.instance_of(str),
            validators.matches_re(r"^[0-9A-Za-z ]+$"),
        ],
    )


@define(kw_only=True)
class Aircraft(AircraftKey):
    # Classes

    @define(kw_only=True)
    class Named:
        @define(kw_only=True)
        class Name:
            label: str = field(
                validator=[
                    validators.instance_of(str),
                    validators.matches_re(r"^[0-9A-Za-z -]+$"),
                ],
            )

        start: PartialDate | None = field(
            converter=to_partial_date,
            default="",
            validator=[
                validators.instance_of(PartialDate | None),
            ],
        )

        finish: PartialDate | None = field(
            converter=to_partial_date,
            default="",
            validator=[
                validators.instance_of(PartialDate | None),
            ],
        )

        name: Name = field(
            validator=[
                validators.instance_of(Name),
            ],
        )

    @define(kw_only=True)
    class OperatedBy:
        @define(kw_only=True)
        class Operator:
            label: str = field(
                validator=[
                    validators.instance_of(str),
                    validators.matches_re(r"^[0-9A-Za-z -]+$"),
                ],
            )

            airline: AirlineKey = field(
                validator=[
                    validators.instance_of(AirlineKey),
                ],
            )

        start: PartialDate | None = field(
            converter=to_partial_date,
            default="",
            validator=[
                validators.instance_of(PartialDate | None),
            ],
        )

        finish: PartialDate | None = field(
            converter=to_partial_date,
            default="",
            validator=[
                validators.instance_of(PartialDate | None),
            ],
        )

        operator: Operator = field(
            validator=[
                validators.instance_of(Operator),
            ],
        )

    @define(kw_only=True)
    class RegisteredAs:
        @define(kw_only=True)
        class Registration:
            label: str = field(
                validator=[
                    validators.instance_of(str),
                    validators.matches_re(r"^[0-9A-Za-z -]+$"),
                ],
            )

        start: PartialDate | None = field(
            converter=to_partial_date,
            default="",
            validator=[
                validators.instance_of(PartialDate | None),
            ],
        )

        finish: PartialDate | None = field(
            converter=to_partial_date,
            default="",
            validator=[
                validators.instance_of(PartialDate | None),
            ],
        )

        registration: Registration = field(
            validator=[
                validators.instance_of(Registration),
            ],
        )

    @define(kw_only=True)
    class Typed:
        @define(kw_only=True)
        class Type:
            label: str = field(
                validator=[
                    validators.instance_of(str),
                    validators.matches_re(r"^[0-9A-Za-z -]+$"),
                ],
            )

        start: PartialDate | None = field(
            converter=to_partial_date,
            default="",
            validator=[
                validators.instance_of(PartialDate | None),
            ],
        )

        finish: PartialDate | None = field(
            converter=to_partial_date,
            default="",
            validator=[
                validators.instance_of(PartialDate | None),
            ],
        )

        type: Type = field(
            validator=[
                validators.instance_of(Type),
            ],
        )

    # Fields

    name: list[Named] = field(
        validator=[
            validators.instance_of(list),
        ],
    )

    operator: list[OperatedBy] = field(
        validator=[
            validators.instance_of(list),
        ],
    )

    registration: list[RegisteredAs] = field(
        validator=[
            validators.instance_of(list),
        ],
    )

    type: list[Typed] = field(
        validator=[
            validators.instance_of(list),
        ],
    )



from collections.abc import Generator
from functools import cached_property
from typing import Any

import rdflib
import requests

from illallangi.rdf.client.aircraft import AircraftMixin
from illallangi.rdf.client.airline import AirlineMixin
from illallangi.rdf.client.airport import AirportMixin
from illallangi.rdf.client.course import CourseMixin
from illallangi.rdf.client.manufacturer import ManufacturerMixin
from illallangi.rdf.client.residence import ResidenceMixin


class RDFClient(
    AircraftMixin,
    AirlineMixin,
    AirportMixin,
    CourseMixin,
    ManufacturerMixin,
    ResidenceMixin,
):
    def __init__(
        self,
        github_file_path: str,
        github_repo_name: str,
        github_repo_owner: str,
        github_token: str,
    ) -> None:
        self.github_file_path = github_file_path
        self.github_repo_name = github_repo_name
        self.github_repo_owner = github_repo_owner
        self.github_token = github_token

    @cached_property
    def graph(self) -> rdflib.Graph:
        url = f"https://api.github.com/repos/{self.github_repo_owner}/{self.github_repo_name}/contents/{self.github_file_path}"
        headers = {
            "Authorization": f"token {self.github_token}",
            "Accept": "application/vnd.github.v3.raw",
        }

        response = requests.get(
            url,
            headers=headers,
            timeout=5,
        )

        response.raise_for_status()

        return rdflib.Graph().parse(data=response.content)

    def query(
        self,
        query: str,
    ) -> Generator[dict[str, str | None], Any, None]:
        results = self.graph.query(
            query,
        )

        for result in results.bindings:
            yield {
                k: v
                for k, v in {
                    str(key): str(result[str(key)]) if str(key) in result else None
                    for key in results.vars
                }.items()
                if v is not None and v not in ["", "None"]
            }



from collections.abc import Generator
from typing import Any

from illallangi.rdf.models import Manufacturer


class ManufacturerMixin:
    def get_manufacturers(
        self,
        *_args: list[Any],
        **_kwargs: dict[str, Any],
    ) -> Generator[Manufacturer, Any, list | None]:
        query = """
    SELECT
        ?label
    WHERE {{
        ?href a ic:manufacturer .
        ?href rdfs:label ?label .
    }}
    """
        for manufacturer in self.query(
            query=query,
        ):
            yield Manufacturer(
                **manufacturer,
            )



from collections.abc import Generator
from typing import Any

from illallangi.rdf.models import (
    Aircraft,
    AirlineKey,
    ManufacturerKey,
)


class AircraftMixin:
    def get_aircraft(
        self,
        *_args: list[Any],
        **_kwargs: dict[str, Any],
    ) -> Generator[Aircraft, Any, list | None]:
        query = """
    PREFIX ic: <http://data.coley.au/rdf/class#>
    PREFIX ip: <http://data.coley.au/rdf/property#>
    SELECT
        ?manufacturer__label
        ?msn
    WHERE {{
        ?href a ic:aircraft .
        ?href ip:aircraftManufacturer [
            rdfs:label ?manufacturer__label ;
        ] .
        ?href ip:aircraftMsn ?msn .
    }}
    """
        for aircraft in self.query(
            query=query,
        ):
            yield Aircraft(
                **{
                    **{k: v for k, v in aircraft.items() if "__" not in k},
                    "manufacturer": ManufacturerKey(
                        label=str(aircraft["manufacturer__label"])
                    ),
                    "name": list(
                        self.get_aircraft_named(
                            **aircraft,
                        ),
                    ),
                    "operator": list(
                        self.get_aircraft_operated_by(
                            **aircraft,
                        ),
                    ),
                    "registration": list(
                        self.get_aircraft_registered_as(
                            **aircraft,
                        ),
                    ),
                    "type": list(
                        self.get_aircraft_typed_as(
                            **aircraft,
                        ),
                    ),
                },
            )

    def get_aircraft_named(
        self,
        *_args: list[Any],
        msn: str,
        manufacturer__label: str,
        **_kwargs: dict[str, Any],
    ) -> Generator[str, Any, list | None]:
        query = f"""
    PREFIX ic: <http://data.coley.au/rdf/class#>
    PREFIX ip: <http://data.coley.au/rdf/property#>
    PREFIX schema: <http://schema.org/>
    SELECT
        ?coverage
        ?aircraft_name__label
    WHERE {{
        ?href a ic:aircraft .
        ?href ip:aircraftMsn {msn} .
        ?href ip:aircraftManufacturer [
            rdfs:label "{manufacturer__label}" ;
        ] .
        ?href ip:aircraftNamed [
            ip:aircraftName [
                rdfs:label ?aircraft_name__label ;
            ] ;
            schema:temporalCoverage ?coverage ;
        ] .
    }}
    """
        for aircraft_name in self.query(
            query=query,
        ):
            yield Aircraft.Named(
                start=aircraft_name["coverage"].split("/")[0],
                finish=aircraft_name["coverage"].split("/")[1]
                if "/" in aircraft_name["coverage"]
                else None,
                name=Aircraft.Named.Name(
                    label=aircraft_name["aircraft_name__label"],
                ),
            )

    def get_aircraft_operated_by(
        self,
        *_args: list[Any],
        msn: str,
        manufacturer__label: str,
        **_kwargs: dict[str, Any],
    ) -> Generator[str, Any, list | None]:
        query = f"""
    PREFIX ic: <http://data.coley.au/rdf/class#>
    PREFIX ip: <http://data.coley.au/rdf/property#>
    PREFIX schema: <http://schema.org/>
    SELECT
        ?coverage
        ?aircraft_operator__label
        ?aircraft_operator__iata_code
    WHERE {{
        ?href a ic:aircraft .
        ?href ip:aircraftMsn {msn} .
        ?href ip:aircraftManufacturer [
            rdfs:label "{manufacturer__label}" ;
        ] .
        ?href ip:aircraftOperatedBy [
            ip:aircraftOperator [
                rdfs:label ?aircraft_operator__label ;
                ip:airline [
                    ip:airlineIataCode ?aircraft_operator__iata_code ;
                ] ;
            ] ;
            schema:temporalCoverage ?coverage ;
        ] .
    }}
    """
        for aircraft_operator in self.query(
            query=query,
        ):
            yield Aircraft.OperatedBy(
                start=aircraft_operator["coverage"].split("/")[0],
                finish=aircraft_operator["coverage"].split("/")[1]
                if "/" in aircraft_operator["coverage"]
                else None,
                operator=Aircraft.OperatedBy.Operator(
                    label=aircraft_operator["aircraft_operator__label"],
                    airline=AirlineKey(
                        iata=str(aircraft_operator["aircraft_operator__iata_code"]),
                    ),
                ),
            )

    def get_aircraft_registered_as(
        self,
        *_args: list[Any],
        msn: str,
        manufacturer__label: str,
        **_kwargs: dict[str, Any],
    ) -> Generator[str, Any, list | None]:
        query = f"""
    PREFIX ic: <http://data.coley.au/rdf/class#>
    PREFIX ip: <http://data.coley.au/rdf/property#>
    PREFIX schema: <http://schema.org/>
    SELECT
        ?coverage
        ?aircraft_registration__label
    WHERE {{
        ?href a ic:aircraft .
        ?href ip:aircraftMsn {msn} .
        ?href ip:aircraftManufacturer [
            rdfs:label "{manufacturer__label}" ;
        ] .
        ?href ip:aircraftRegisteredAs [
            ip:aircraftRegistration [
                rdfs:label ?aircraft_registration__label ;
            ] ;
            schema:temporalCoverage ?coverage ;
        ] .
    }}
    """
        for aircraft_registration in self.query(
            query=query,
        ):
            yield Aircraft.RegisteredAs(
                start=aircraft_registration["coverage"].split("/")[0],
                finish=aircraft_registration["coverage"].split("/")[1]
                if "/" in aircraft_registration["coverage"]
                else None,
                registration=Aircraft.RegisteredAs.Registration(
                    label=aircraft_registration["aircraft_registration__label"],
                ),
            )

    def get_aircraft_typed_as(
        self,
        *_args: list[Any],
        msn: str,
        manufacturer__label: str,
        **_kwargs: dict[str, Any],
    ) -> Generator[str, Any, list | None]:
        query = f"""
    PREFIX ic: <http://data.coley.au/rdf/class#>
    PREFIX ip: <http://data.coley.au/rdf/property#>
    PREFIX schema: <http://schema.org/>
    SELECT
        ?coverage
        ?aircraft_type__label
    WHERE {{
        ?href a ic:aircraft .
        ?href ip:aircraftMsn {msn} .
        ?href ip:aircraftManufacturer [
            rdfs:label "{manufacturer__label}" ;
        ] .
        ?href ip:aircraftTypedAs [
            ip:aircraftType [
                rdfs:label ?aircraft_type__label ;
            ] ;
            schema:temporalCoverage ?coverage ;
        ] .
    }}
    """
        for aircraft_type in self.query(
            query=query,
        ):
            yield Aircraft.Typed(
                start=aircraft_type["coverage"].split("/")[0],
                finish=aircraft_type["coverage"].split("/")[1]
                if "/" in aircraft_type["coverage"]
                else None,
                type=Aircraft.Typed.Type(
                    label=aircraft_type["aircraft_type__label"],
                ),
            )



from partial_date import PartialDate


def to_partial_date(
    value: str | PartialDate,
) -> PartialDate:
    if value is None:
        return None
    if isinstance(value, PartialDate):
        return value
    if isinstance(value, str):
        if value in ["", "Unknown"]:
            return None
        return PartialDate(value)
    msg = f"Invalid PartialDate: {value}"
    raise ValueError(msg)



from attrs import define, field, validators


@define(kw_only=True)
class ManufacturerKey:
    # Natural Keys

    label: str = field(
        validator=[
            validators.instance_of(str),
            validators.matches_re(r"^[0-9A-Za-z -]+$"),
        ],
    )


@define(kw_only=True)
class Manufacturer(ManufacturerKey):
    pass
